import numpy as np
import configparser
from pathlib import Path
from functools import cached_property
import aquila_borg as borg


class BORGConfig:
    """
    This is a Hades/BORG ini file parser.

    It helps at parsing the ini file and instancing Aquila_borg objects
    from the instructions contained within. It allows the user to
    safely reprocess MCMC chains with exactly the same setting.

    Arguments:
      * path (Path): path to the ini file

    Attributes:
      Ndata (tuple): A tuple containing the number of data points in each dimension.
      config (ConfigParser): The configuration parser object.
      cosmology (CosmologicalParameters): The cosmological parameters parsed from the ini file.
      ic_box (BoxModel): The initial conditions box model parsed from the ini file.
      likelihood_info (LikelihoodInfo): The likelihood information parsed from the ini file.
      likelihood (Likelihood): The likelihood object created based on the ini file.
      density_sampler (HMCDensitySampler): The density sampler object created based on the ini file.

    Methods:
      build_gravity_model(): Builds and returns the gravity model chain and dynamic model.
    """

    def __init__(self, path: Path):
        self._parser = configparser.ConfigParser()
        self._info = None
        self._parser.read(path)

    @cached_property
    def N(self):
        """
        Retrieve the N values from the configuration file.

        Returns:
            tuple: A tuple containing three integer values representing N.
        """
        if "system" not in self._parser:
            raise ValueError("No system section in the configuration file")
        if not all(f"N{a}" in self._parser["system"] for a in range(3)):
            raise ValueError("Missing N values in the configuration file")
        return tuple(int(self._parser["system"][f"N{a}"]) for a in range(3))

    @cached_property
    def Ndata(self):
        """
        Retrieve the Ndata values from the configuration file.

        Returns:
            tuple: A tuple containing three integer values representing Ndata.
        """
        if "system" not in self._parser:
            raise ValueError("No system section in the configuration file")
        # Defaulting to N if Ndata is not present
        if not all(f"Ndata{i}" in self._parser["system"] for i in range(3)):
            return self.N
        return tuple(int(self._parser["system"][f"Ndata{i}"]) for i in range(3))

    @property
    def config(self):
        """
        Returns the configuration parser object used to parse the ini file.

        Returns:
            ConfigParser: The parser configuration object.
        """
        return self._parser

    @cached_property
    def cosmology(self):
        """
        Generates and returns a CosmologicalParameters object with attributes set
        based on the 'cosmology' section of the ini file.

        Returns:
            aquila_borg.cosmo.CosmologicalParameters: An instance of CosmologicalParameters
        """

        cpar = borg.cosmo.CosmologicalParameters()
        for c in self._parser["cosmology"]:
            if hasattr(cpar, c):
                setattr(cpar, c, float(self._parser["cosmology"][c]))
        return cpar

    @cached_property
    def ic_box(self):
        """
        Create and return a BoxModel object with parameters parsed from the configuration file.
        This box model corresponds to the primary box for initial conditions.

        Returns:
            borg.forward.BoxModel: An instance of BoxModel with the parsed configuration parameters.
        """
        bbox = borg.forward.BoxModel()
        bbox.N = tuple(int(self._parser["system"][f"N{a}"]) for a in range(3))
        bbox.L = tuple(float(self._parser["system"][f"L{a}"]) for a in range(3))
        bbox.xmin = tuple(float(self._parser["system"][f"corner{a}"]) for a in range(3))
        return bbox

    @property
    def likelihood_info(self):
        """
        Retrieves or initializes the likelihood information.
        The LikelihoodInfo object is cached and populatesd with values
        from the parser's "likelihood" section.

        Returns:
            LikelihoodInfo: The likelihood information object.
        """

        if self._info:
            return self._info
        self._info = borg.likelihood.LikelihoodInfo()
        for c in self._parser["likelihood"]:
            if hasattr(self._info, c):
                setattr(self._info, c, float(self._parser["likelihood"][c]))
        return self._info

    @cached_property
    def likelihood(self):
        """
        Creates and returns a new likelihood object using the likelihoods registered in
        the aquila_borg library.

        This method initializes a new likelihood object based on the configuration
        specified in the parser and the likelihood information.

        Returns:
            aquila_borg.likelihood.Likelihood3d: A new likelihood object.
        """

        likelihood = borg.likelihood.newLikelihood(
            self._parser["hades"]["likelihood"], self.likelihood_info
        )
        return likelihood

    @cached_property
    def density_sampler(self):
        algo_name = self._parser["hades"]["algorithm"]
        if algo_name == "HMC":
            return borg.samplers.HMCDensitySampler
        elif algo_name == "AO-HMC":
            return borg.samplers.AOHMCDensitySampler
        else:
            raise ValueError(f"Unknown algorithm {algo_name}")

    def build_gravity_model(self):
        """
        Builds the gravity model chain based on the configuration provided in the parser.
        This method reads the gravity models specified in the ini file, constructs a chain of forward models,
        and returns the chain along with the model marked "dynamic" if present.

        Returns:
            tuple: A tuple containing:
                - chain (borg.forward.ChainForwardModel): The chain of forward models.
                - dynamic_model (borg.forward.ParticleBasedForwardModel or None): The dynamic model if present, otherwise None.
        """

        dynamic_model = None

        model_list = self._parser["gravity"]["models"].split(",")

        chain = borg.forward.ChainForwardModel(self.ic_box)
        bbox = self.ic_box
        for a, i in enumerate(model_list):
            opts = dict(self._parser[f"gravity_chain_{a}"])
            m = borg.forward.models.newModel(i, bbox, opts=opts)
            bbox = m.getOutputBoxModel()
            if "name" in opts:
                m.setName(opts["name"])
                if opts["name"] == "dynamics":
                    dynamics_model = m
            chain @= m
            if dynamic_model is None and isinstance(
                m, borg.forward.ParticleBasedForwardModel
            ):
                dynamic_model = m

        return chain, dynamic_model

    def fill_catalog(
        self, state: borg.likelihood.MarkovState, idcat: int, catalog: dict
    ):
        """
        Fill the markov state with catalog information as indicated by the configuration file.

        """
        state.newGalaxyClusteringElement(f"galaxy_catalog_{idcat}")
        bias_name = f"galaxy_bias_{idcat}"
        if "bias" in catalog:
            bias = np.array(list(map(float, catalog["bias"].split(","))))
            state.resizeArray1d(bias_name, bias.size)
            state[bias_name][:] = bias

        win3d = state[f"galaxy_sel_window_{idcat}"]
        #borg.data.(catalog["window"], win3d)
        # state[f"galaxy_catalog_{idcat}"] = catalog
        # state[f"galaxy_catalog_{idcat}_N"] = len(catalog)
        # for i, gal in enumerate(catalog):
        #     for k, v in gal.items():
        #         state[f"galaxy_catalog_{idcat}_{i}_{k}"] = v
